from pymobiledevice3.services.screenshot import ScreenshotService

PNG_HEADER = b'\x89\x50\x4e\x47\x0d\x0a\x1a\x0a'


def test_screenshot(lockdown):
    """
    Test that taking a screenshot returns a PNG.
    :param pymobiledevice3.lockdown.LockdownClient lockdown: Lockdown client.
    """
    screenshot_taker = ScreenshotService(lockdown)
    screenshot = screenshot_taker.take_screenshot()
    assert screenshot.startswith(PNG_HEADER)



import logging
import plistlib
from uuid import uuid4

import asn1
import requests

from pymobiledevice3.restore.img4 import img4_get_component_tag

TSS_CLIENT_VERSION_STRING = 'libauthinstall-776.60.1'


class TSSRequest:
    def __init__(self):
        self._request = {
            '@BBTicket': True,
            '@HostPlatformInfo': 'mac',
            '@VersionInfo': TSS_CLIENT_VERSION_STRING,
            '@UUID': str(uuid4()).upper(),
        }

    def add_common_flags(self, parameters, overrides=None):
        keys = ('ApECID', 'UniqueBuildID', 'ApChipID', 'ApBoardID', 'ApSecurityDomain')
        for k in keys:
            if k in parameters:
                self._request[k] = parameters[k]
        if overrides is not None:
            self._request.update(overrides)

    def add_vinyl_tags(self, parameters, overrides=None):
        self._request['@BBTicket'] = True
        self._request['@eUICC,Ticket'] = True

        keys = ('eUICC,ChipID', 'eUICC,EID', 'eUICC,RootKeyIdentifier')
        for k in keys:
            if k in parameters:
                self._request[k] = parameters[k]

        # set Nonce for eUICC,Gold component
        node = parameters.get('EUICCGoldNonce')
        if node is not None:
            n = self._request['eUICC,Gold']
            n['Nonce'] = node

        # set Nonce for eUICC,Main component
        node = parameters.get('EUICCMainNonce')
        if node is not None:
            n = self._request['eUICC,Main']
            n['Nonce'] = node

        if overrides is not None:
            self._request.update(overrides)

    @staticmethod
    def apply_restore_request_rules(tss_entry: dict, parameters: dict, rules: list):
        for rule in rules:
            conditions_fulfilled = True
            conditions = rule['Conditions']
            for key, value in conditions.items():
                if not conditions_fulfilled:
                    break

                if key == 'ApRawProductionMode':
                    value2 = parameters.get('ApProductionMode')
                elif key == 'ApCurrentProductionMode':
                    value2 = parameters.get('ApProductionMode')
                elif key == 'ApRawSecurityMode':
                    value2 = parameters.get('ApSecurityMode')
                elif key == 'ApRequiresImage4':
                    value2 = parameters.get('ApSupportsImg4')
                elif key == 'ApDemotionPolicyOverride':
                    value2 = parameters.get('DemotionPolicy')
                elif key == 'ApInRomDFU':
                    value2 = parameters.get('ApInRomDFU')
                else:
                    logging.error(f'Unhandled condition {key} while parsing RestoreRequestRules')
                    value2 = None

                if value2:
                    conditions_fulfilled = value == value2
                else:
                    conditions_fulfilled = False

            if not conditions_fulfilled:
                continue

            actions = rule['Actions']
            for key, value in actions.items():
                if value != 255:
                    value2 = tss_entry.get(key)
                    if value2:
                        tss_entry.pop(key)
                    logging.debug(f'DEBUG: Adding {key}={value} to TSS entry')
                    tss_entry[key] = value
        return tss_entry

    def add_ap_tags(self, parameters, overrides=None):
        """ loop over components from build manifest """

        manifest_node = parameters['Manifest']

        # add components to request
        skipped_keys = ('BasebandFirmware', 'SE,UpdatePayload', 'BaseSystem', 'Diags',)
        for key, manifest_entry in manifest_node.items():
            if key in skipped_keys:
                continue

            if parameters.get('_OnlyFWComponents', False):
                if not manifest_node.get('Trusted', False):
                    logging.debug(f'skipping {key} as it is not trusted')
                    continue
                info = manifest_node['Info']
                if not info['IsFirmwarePayload'] and not info['IsSecondaryFirmwarePayload'] and not info[
                    'IsFUDFirmware']:
                    logging.debug(f'skipping {key} as it is neither firmware nor secondary nor FUD firmware payload')
                    continue

            # copy this entry
            tss_entry = dict(manifest_entry)

            # remove obsolete Info node
            tss_entry.pop('Info')

            # handle RestoreRequestRules
            if 'Info' in manifest_entry and 'RestoreRequestRules' in manifest_entry['Info']:
                rules = manifest_entry['Info']['RestoreRequestRules']
                if rules:
                    logging.debug(f'Applying restore request rules for entry {key}')
                    tss_entry = self.apply_restore_request_rules(tss_entry, parameters, rules)

            # Make sure we have a Digest key for Trusted items even if empty
            node = manifest_entry.get('Trusted', False)
            if node:
                if manifest_entry.get('Digest') is None:
                    tss_entry['Digest'] = bytes()

            self._request[key] = tss_entry

        if overrides is not None:
            self._request.update(overrides)

    def add_ap_img3_tags(self, parameters):
        if 'ApNonce' in parameters:
            self._request['ApNonce'] = parameters['ApNonce']
        self._request['@APTicket'] = True

    def add_ap_img4_tags(self, parameters):
        keys_to_copy = (
            'ApNonce', 'Ap,OSLongVersion', 'ApSecurityMode', 'ApProductionMode', 'ApSepNonce',
            'PearlCertificationRootPub')
        for k in keys_to_copy:
            if k in parameters:
                v = parameters[k]
                if k == 'ApSepNonce':
                    k = 'SepNonce'
                self._request[k] = v

        self._request['@ApImg4Ticket'] = True

    def add_baseband_tags(self, parameters, overrides=None):
        keys_to_copy = (
            'BbChipID', 'BbProvisioningManifestKeyHash', 'BbActivationManifestKeyHash', 'BbCalibrationManifestKeyHash',
            'BbFactoryActivationManifestKeyHash', 'BbFDRSecurityKeyHash', 'BbSkeyId', 'BbNonce', '@BBTicket',
            'BbGoldCertId', 'BbSNUM',)

        for k in keys_to_copy:
            if k in parameters:
                self._request[k] = parameters[k]

        bb_chip_id = parameters['BbChipID']
        bb_cert_id = parameters['BbGoldCertId']

        bbfwdict = dict(parameters['Manifest']['BasebandFirmware'])
        bbfwdict.pop('Info')

        if bb_chip_id == 0x68:
            # depending on the BasebandCertId remove certain nodes
            if bb_cert_id in (0x26F3FACC, 0x5CF2EC4E, 0x8399785A):
                bbfwdict.pop('PSI2-PartialDigest')
                bbfwdict.pop('RestorePSI2-PartialDigest')
            else:
                bbfwdict.pop('PSI-PartialDigest')
                bbfwdict.pop('RestorePSI-PartialDigest')

        self._request['BasebandFirmware'] = bbfwdict

        if overrides:
            self._request.update(overrides)

    def img4_create_local_manifest(self):
        p = asn1.Encoder()
        p.start()

        p.write(b'MANP', asn1.Numbers.IA5String)
        p.enter(asn1.Numbers.Set)

        p.write(b'BORD', asn1.Numbers.IA5String)
        p.write(self._request['ApBoardID'], asn1.Numbers.Integer)

        p.write(b'CEPO', asn1.Numbers.IA5String)
        p.write(0, asn1.Numbers.Integer)

        p.write(b'CHIP', asn1.Numbers.IA5String)
        p.write(self._request['ApChipID'], asn1.Numbers.Integer)

        p.write(b'CPRO', asn1.Numbers.IA5String)
        p.write(self._request['ApProductionMode'], asn1.Numbers.Integer)

        p.write(b'CSEC', asn1.Numbers.IA5String)
        p.write(0, asn1.Numbers.Integer)

        p.write(b'SDOM', asn1.Numbers.IA5String)
        p.write(self._request['ApSecurityDomain'], asn1.Numbers.Integer)

        p.leave()

        # now write the components
        for k, v in self._request.items():
            if isinstance(v, dict):
                comp = img4_get_component_tag(k)
                if comp is None:
                    raise Exception(f'Unhandled component {k} - can\'t create manifest')
                logging.debug(f'found component {comp} ({k})')

        # write manifest body header
        p.write(b'MANB', asn1.Numbers.IA5String)
        p.enter(asn1.Numbers.Set)
        p.leave()

        # write header values
        p.write(b'IM4M', asn1.Numbers.IA5String)
        p.write(0, asn1.Numbers.Integer)

        return p.output()

    def request_send(self):
        for retry in range(15):
            headers = {
                'Cache-Control': 'no-cache',
                'Content-type': 'text/xml; charset="utf-8"',
                'User-Agent': 'InetURL/1.0',
                'Expect': '',
            }

            logging.info(f'Sending TSS request attempt {retry}')
            r = requests.post('http://gs.apple.com/TSS/controller?action=2', headers=headers,
                              data=plistlib.dumps(self._request), verify=False)

            if b'MESSAGE=SUCCESS' in r.content:
                logging.info('response successfully received')

            message = r.content.split(b'MESSAGE=', 1)[1].split(b'&', 1)[0].decode()
            if message != 'SUCCESS':
                raise Exception(f'server replied: {message}')

            return plistlib.loads(r.content.split(b'REQUEST_STRING=', 1)[1])



def img4_get_component_tag(compname):
    component_tags = {
        'ACIBT': b'acib',
        'ACIBTLPEM': b'lpbt',
        'ACIWIFI': b'aciw',
        'Alamo': b'almo',
        'ANE': b'anef',
        'ANS': b'ansf',
        'AOP': b'aopf',
        'Ap,AudioAccessibilityBootChime': b'auac',
        'Ap,AudioBootChime': b'aubt',
        'Ap,AudioPowerAttachChime': b'aupr',
        'Ap,CIO': b'ciof',
        'Ap,HapticAssets': b'hpas',
        'Ap,LocalBoot': b'lobo',
        'Ap,LocalPolicy': b'lpol',
        'Ap,NextStageIM4MHash': b'nsih',
        'Ap,RecoveryOSPolicyNonceHash': b'ronh',
        'Ap,RestoreCIO': b'rcio',
        'Ap,RestoreTMU': b'rtmu',
        'Ap,Scorpius': b'scpf',
        'Ap,TMU': b'tmuf',
        'Ap,VolumeUUID': b'vuid',
        'AppleLogo': b'logo',
        'AudioCodecFirmware': b'acfw',
        'AVE': b'avef',
        'BatteryCharging': b'glyC',
        'BatteryCharging0': b'chg0',
        'BatteryCharging1': b'chg1',
        'BatteryFull': b'batF',
        'BatteryLow0': b'bat0',
        'BatteryLow1': b'bat1',
        'BatteryPlugin': b'glyP',
        'CFELoader': b'cfel',
        'Dali': b'dali',
        'DCP': b'dcpf',
        'DeviceTree': b'dtre',
        'Diags': b'diag',
        'EngineeringTrustCache': b'dtrs',
        'ExtDCP': b'edcp',
        'ftap': b'ftap',
        'ftsp': b'ftsp',
        'GFX': b'gfxf',
        'Hamm': b'hamf',
        'Homer': b'homr',
        'iBEC': b'ibec',
        'iBoot': b'ibot',
        'iBootData': b'ibdt',
        'iBootTest': b'itst',
        'iBSS': b'ibss',
        'InputDevice': b'ipdf',
        'ISP': b'ispf',
        'KernelCache': b'krnl',
        'LeapHaptics': b'lphp',
        'Liquid': b'liqd',
        'LLB': b'illb',
        'LoadableTrustCache': b'ltrs',
        'LowPowerWallet0': b'lpw0',
        'LowPowerWallet1': b'lpw1',
        'LowPowerWallet2': b'lpw2',
        'MacEFI': b'mefi',
        'Multitouch': b'mtfw',
        'NeedService': b'nsrv',
        'OS': b'OS\0\0',
        'OSRamdisk': b'osrd',
        'PersonalizedDMG': b'pdmg',
        'PEHammer': b'hmmr',
        'PERTOS': b'pert',
        'PHLEET': b'phlt',
        'PMP': b'pmpf',
        'RBM': b'rmbt',
        'Rap,SoftwareBinaryDsp1': b'sbd1',
        'Rap,RTKitOS': b'rkos',
        'Rap,RestoreRTKitOS': b'rrko',
        'RecoveryMode': b'recm',
        'RestoreANS': b'rans',
        'RestoreDCP': b'rdcp',
        'RestoreDeviceTree': b'rdtr',
        'RestoreExtDCP': b'recp',
        'RestoreKernelCache': b'rkrn',
        'RestoreLogo': b'rlgo',
        'RestoreRamDisk': b'rdsk',
        'RestoreSEP': b'rsep',
        'RestoreTrustCache': b'rtsc',
        'rfta': b'rfta',
        'rfts': b'rfts',
        'RTP': b'rtpf',
        'SCE': b'scef',
        'SCE1Firmware': b'sc1f',
        'SEP': b'sepi',
        'SIO': b'siof',
        'StaticTrustCache': b'trst',
        'SystemLocker': b'lckr',
        'WCHFirmwareUpdater': b'wchf',
    }

    return component_tags.get(compname)



import click

from pymobiledevice3.cli.cli_common import Command, print_json
from pymobiledevice3.services.misagent import MisagentService


@click.group()
def cli():
    """ privision cli """
    pass


@cli.group()
def provision():
    """ privision options """
    pass


@provision.command('install', cls=Command)
@click.argument('profile', type=click.File('rb'))
def provision_install(lockdown, profile):
    """ install a provision profile (.mobileprovision file) """
    MisagentService(lockdown=lockdown).install(profile)


@provision.command('remove', cls=Command)
@click.argument('profile_id')
def provision_install(lockdown, profile_id):
    """ remove a provision profile """
    MisagentService(lockdown=lockdown).remove(profile_id)


@provision.command('list', cls=Command)
@click.option('--color/--no-color', default=True)
def provision_list(lockdown, color):
    """ list installed provision profiles """
    print_json(MisagentService(lockdown=lockdown).copy_all(), colored=color)



import logging
import plistlib
import time
import zipfile
from io import BytesIO

import asn1
from usb import USBError

from pymobiledevice3.exceptions import PyMobileDevice3Exception
from pymobiledevice3.irecv import IRecv
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.restore.tss import TSSRequest
from pymobiledevice3.services.preboard import PreboardService


class RestoreService:
    def __init__(self, lockdown: LockdownClient, ipsw: BytesIO, tss=None):
        self.ipsw = zipfile.ZipFile(ipsw)
        self.tss = tss

        self._lockdown = lockdown
        self.hardware_model = self._lockdown.all_values['HardwareModel'].lower()
        self.ecid = self._lockdown.all_values['UniqueChipID']
        self.build_version = self._lockdown.all_values['BuildVersion']
        self.build_major = int(self.build_version[:2])
        self.is_image4_supported = self._lockdown.get_value(key='Image4Supported')

        if not self.is_image4_supported:
            raise NotImplementedError('is_image4_supported is False')

        logging.info(f'connected device: <ecid: {self.ecid} chipset: {self.hardware_model} build: {self.build_version} '
                     f'image4-support: {self.is_image4_supported}>')

        self.build_identity = None

        logging.debug('scanning BuildManifest.plist for the correct BuildIdentity')
        self.build_manifest = plistlib.loads(self.ipsw.read('BuildManifest.plist'))
        for build_identity in self.build_manifest['BuildIdentities']:
            device_class = build_identity['Info']['DeviceClass'].lower()
            restore_behavior = build_identity['Info']['RestoreBehavior']
            logging.debug(f'iterating: class: {device_class} behavior: {restore_behavior}')
            if (device_class == self.hardware_model) and (restore_behavior == 'Update'):
                self.build_identity = build_identity
                break

        if self.build_identity is None:
            raise PyMobileDevice3Exception('failed to find the correct BuildIdentity from the BuildManifest.plist')

        self.irecv = None  # type: IRecv

    def reconnect(self):
        logging.debug('Waiting for device to reconnect in Recovery mode...')
        self.irecv = IRecv(ecid=self.ecid)

    def get_preboard_manifest(self):
        overrides = {
            '@APTicket': True,
            'ApProductionMode': 0,
            'ApSecurityDomain': 0,
        }

        parameters = {
            'ApProductionMode': False,
            'ApSecurityMode': False,
            'ApSupportsImg4': True,
        }

        keys_to_copy = ('UniqueBuildID', 'ApChipID', 'ApBoardID', 'ApSecurityDomain', 'Ap,OSLongVersion', 'BMU,BoardID',
                        'BMU,ChipID', 'BbChipID', 'BbProvisioningManifestKeyHash', 'BbActivationManifestKeyHash',
                        'BbCalibrationManifestKeyHash', 'BbFactoryActivationManifestKeyHash', 'BbFDRSecurityKeyHash',
                        'BbSkeyId', 'SE,ChipID', 'Savage,ChipID', 'Savage,PatchEpoch', 'Yonkers,BoardID',
                        'Yonkers,ChipID', 'Yonkers,PatchEpoch', 'Rap,BoardID', 'Rap,ChipID', 'Rap,SecurityDomain',
                        'eUICC,ChipID', 'PearlCertificationRootPub', 'Manifest',)
        for k in keys_to_copy:
            try:
                v = self.build_identity[k]
                if isinstance(v, str) and v.startswith('0x'):
                    v = int(v, 16)
                parameters[k] = v
            except KeyError:
                pass

        tss = TSSRequest()
        tss.add_common_flags(parameters, overrides)

        parameters['_OnlyFWComponents'] = True

        tss.add_ap_tags(parameters)

        return tss.img4_create_local_manifest()

    def normal_handle_create_stashbag(self, manifest):
        preboard = PreboardService(self._lockdown)
        return preboard.create_stashbag(manifest)

    def get_ap_nonce(self):
        return self._lockdown.get_value(key='ApNonce')

    def get_sep_nonce(self):
        return self._lockdown.get_value(key='SEPNonce')

    def tss_parameters_add_from_manifest(self, parameters):
        keys_to_copy = (
            'UniqueBuildID', 'Ap,OSLongVersion', 'ApChipID', 'ApBoardID', 'ApSecurityDomain',
            'BMU,BoardID', 'BMU,ChipID', 'BbChipID', 'BbProvisioningManifestKeyHash', 'BbActivationManifestKeyHash',
            'BbCalibrationManifestKeyHash', 'BbFactoryActivationManifestKeyHash', 'BbFDRSecurityKeyHash', 'BbSkeyId',
            'SE,ChipID', 'Savage,ChipID', 'Savage,PatchEpoch', 'Yonkers,BoardID', 'Yonkers,ChipID',
            'Yonkers,PatchEpoch', 'Rap,BoardID', 'Rap,ChipID', 'Rap,SecurityDomain', 'eUICC,ChipID',
            'PearlCertificationRootPub', 'Manifest')

        for k in keys_to_copy:
            try:
                v = self.build_identity[k]
                if isinstance(v, str) and v.startswith('0x'):
                    v = int(v, 16)
                parameters[k] = v
            except KeyError:
                pass

    def get_preflight_info(self):
        return self._lockdown.get_value(key='FirmwarePreflightInfo')

    def get_tss_response(self):
        # populate parameters
        parameters = dict()

        parameters['ApECID'] = self.ecid
        if self.ap_nonce is not None:
            parameters['ApNonce'] = self.ap_nonce

        sep_nonce = self.get_sep_nonce()
        if sep_nonce is not None:
            parameters['ApSepNonce'] = sep_nonce

        parameters['ApProductionMode'] = True

        if self.is_image4_supported:
            parameters['ApSecurityMode'] = True
            parameters['ApSupportsImg4'] = True
        else:
            parameters['ApSupportsImg4'] = False

        self.tss_parameters_add_from_manifest(parameters)

        tss = TSSRequest()
        tss.add_common_flags(parameters)
        tss.add_ap_tags(parameters)

        # add personalized parameters
        if self.is_image4_supported:
            tss.add_ap_img4_tags(parameters)
        else:
            tss.add_ap_img3_tags(parameters)

        # normal mode; request baseband ticket aswell
        pinfo = self.get_preflight_info()
        if pinfo:
            logging.debug('adding preflight info')

            node = pinfo.get('Nonce')
            if node is not None:
                parameters['BbNonce'] = node

            node = pinfo.get('ChipID')
            if node is not None:
                parameters['BbChipID'] = node

            node = pinfo.get('CertID')
            if node is not None:
                parameters['BbGoldCertId'] = node

            node = pinfo.get('ChipSerialNo')
            if node is not None:
                parameters['BbSNUM'] = node

            tss.add_baseband_tags(parameters)

            euiccchipid = pinfo.get('EUICCChipID')
            if euiccchipid:
                logging.debug('adding EUICCChipID info')
                parameters['eUICC,ChipID'] = euiccchipid

                if euiccchipid >= 5:
                    node = pinfo.get('EUICCCSN')
                    if node is not None:
                        parameters['eUICC,EID'] = node

                    node = pinfo.get('EUICCCertIdentifier')
                    if node is not None:
                        parameters['eUICC,RootKeyIdentifier'] = node

                    node = pinfo.get('EUICCGoldNonce')
                    if node is not None:
                        parameters['EUICCGoldNonce'] = node

                    node = pinfo.get('EUICCMainNonce')
                    if node is not None:
                        parameters['EUICCMainNonce'] = node

                    tss.add_vinyl_tags(parameters)

        # send request and grab response
        return tss.request_send()

    def fetch_tss_record(self):
        stashbag_commit_required = False
        if self._lockdown.all_values['HasSiDP']:
            logging.info('Checking if device requires stashbag...')
            manifest = self.get_preboard_manifest()

            logging.debug('creating stashbag...')
            response = self.normal_handle_create_stashbag(manifest)

            if not response.get('Skip', False):
                raise NotImplementedError('requiring stashbag not yet supported')

        if self.build_major > 8:
            self.ap_nonce = self.get_ap_nonce()
            if self.ap_nonce is None:
                # the first nonce request with older firmware releases can fail and it's OK
                logging.info('NOTE: Unable to get nonce from device')

        self.tss = self.get_tss_response()

        if self.build_major >= 20:
            raise NotImplementedError('not yet supported')

        if stashbag_commit_required:
            raise NotImplementedError('requiring stashbag not yet supported')

        return self.tss

    def tss_response_get_path_by_entry(self, component):
        node = self.tss.get(component)
        if node is not None:
            return node.get('Path')

        return None

    def build_identity_get_component_path(self, component):
        return self.build_identity['Manifest'][component]['Info']['Path']

    def tss_response_get_ap_img4_ticket(self):
        return self.tss.get('ApImg4Ticket')

    def img4_stitch_component(self, name, data, blob):
        logging.info(f'Personalizing IMG4 component {name}...')

        # first we need check if we have to change the tag for the given component
        decoder = asn1.Decoder()
        decoder.start(data)
        decoder.enter()

        decoder.read()
        tag, value = decoder.read()

        component_name_tag = {
            'RestoreKernelCache': b'rkrn',
            'RestoreDeviceTree': b'rdtr',
            'RestoreSEP': b'rsep',
            'RestoreLogo': b'rlgo',
            'RestoreTrustCache': b'rtsc',
            'RestoreDCP': b'rdcp',
            'Ap,RestoreTMU': b'rtmu',
            'Ap,RestoreCIO': b'rcio',
            'Ap,DCP2': b'dcp2',
        }

        logging.debug(f'tag: {tag} {value}')
        if name in component_name_tag:
            logging.debug('Tag found')
            data = data.replace(value.encode(), component_name_tag[name], 1)

        # create element header for the "IMG4" magic
        encoder = asn1.Encoder()
        encoder.start()

        encoder.enter(asn1.Numbers.Sequence)

        # create element header for the "IMG4" magic
        encoder.write(b'IMG4', asn1.Numbers.IA5String)

        decoder = asn1.Decoder()
        decoder.start(data)
        encoder.write(decoder.read()[1], nr=asn1.Numbers.Sequence, typ=asn1.Types.Constructed)

        encoder.enter(0, cls=asn1.Classes.Context)

        decoder = asn1.Decoder()
        decoder.start(blob)
        encoder.write(decoder.read()[1], nr=asn1.Numbers.Sequence, typ=asn1.Types.Constructed)

        encoder.leave()

        encoder.leave()

        return encoder.output()

    def personalize_component(self, name, data):
        blob = self.tss_response_get_ap_img4_ticket()

        # stitch ApImg4Ticket into IMG4 file
        return self.img4_stitch_component(name, data, blob)

    def recovery_send_component(self, name):
        logging.info(f'Sending {name}...')
        path = None
        if self.tss:
            path = self.tss_response_get_path_by_entry(name)

        if path is None:
            logging.debug(f'NOTE: No path for component {name} in TSS, will fetch from build_identity')

        path = self.build_identity_get_component_path(name)
        logging.debug(f'Sending a patched version of: {path}')

        data = self.ipsw.read(path)
        signed_data = self.personalize_component(name, data)

        self.irecv.send_buffer(signed_data)

    def recovery_send_component_and_command(self, name, command):
        self.recovery_send_component(name)
        self.irecv.send_command(command)

    def recovery_send_ibec(self):
        component = 'iBEC'
        self.recovery_send_component(component)
        self.irecv.send_command('go')
        self.irecv.ctrl_transfer(0x21, 1)

    def recovery_send_applelogo(self):
        component = 'RestoreLogo'
        self.recovery_send_component(component)
        self.irecv.send_command('setpicture 4')
        self.irecv.send_command('bgcolor 0 0 0')

    def recovery_send_loaded_by_iboot(self):
        manifest = self.build_identity['Manifest']
        for key, node in manifest.items():
            iboot = node['Info'].get('IsLoadedByiBoot', False)
            iboot_stg1 = node['Info'].get('IsLoadedByiBootStage1', False)

            assert isinstance(iboot, bool)
            assert isinstance(iboot_stg1, bool)

            if iboot and not iboot_stg1:
                logging.debug(f'{key} is loaded by iBoot')
                self.recovery_send_component_and_command(key, 'firmware')

    def recovery_send_ramdisk(self):
        component = 'RestoreRamDisk'
        ramdisk_size = self.irecv.getenv('ramdisk-size')
        logging.info(f'ramdisk-size: {ramdisk_size}')

        self.recovery_send_component(component)
        ramdisk_delay = self.irecv.getenv('ramdisk-delay')
        logging.info(f'ramdisk-delay: {ramdisk_delay}')

        self.irecv.send_command('ramdisk')

        time.sleep(2)

    def recovery_send_kernelcache(self):
        component = 'RestoreKernelCache'

        self.recovery_send_component(component)
        try:
            self.irecv.ctrl_transfer(0x21, 1)
        except USBError:
            pass

        if self.restore_boot_args:
            self.irecv.send_command(f'setenv boot-args {self.restore_boot_args}')

        try:
            self.irecv.send_command('bootx')
        except USBError:
            pass

    def recovery_set_autoboot(self, enable: bool):
        self.irecv.set_autoboot(enable)

    def recovery_enter_restore(self):
        if self.build_major >= 8:
            self.restore_boot_args = 'rd=md0 nand-enable-reformat=1 -progress'
        elif self.build_major >= 20:
            self.restore_boot_args = 'rd=md0 nand-enable-reformat=1 -progress -restore'

        # upload data to make device boot restore mode

        # Recovery Mode Environment:
        build_version = self.irecv.getenv('build-version')
        logging.info(f'iBoot build-version={build_version}')

        build_style = self.irecv.getenv('build-style')
        logging.info(f'iBoot build-style={build_style}')

        radio_error = self.irecv.getenv('radio-error')
        if radio_error:
            radio_error = int(radio_error)
            logging.info(f'radio-error: {radio_error}')
            radio_error_string = self.irecv.getenv('radio-error-string')
            if radio_error_string:
                logging.info(f'radio-error-string: {radio_error_string}')

        self.recovery_set_autoboot(False)

        # send logo and show it
        self.recovery_send_applelogo()

        # send components loaded by iBoot
        self.recovery_send_loaded_by_iboot()

        # send ramdisk and run it
        self.recovery_send_ramdisk()

        # send devicetree and load it
        self.recovery_send_component_and_command('RestoreDeviceTree', 'devicetree')

        if 'RestoreSEP' in self.build_identity['Manifest']:
            # send rsepfirmware and load it
            self.recovery_send_component_and_command('RestoreSEP', 'rsepfirmware')

        self.recovery_send_kernelcache()

    def boot_ramdisk(self):
        if self.tss is None:
            logging.info('fetching TSS record')
            self.fetch_tss_record()

        logging.info('going into Recovery')
        self._lockdown.enter_recovery()
        self.reconnect()
        ecid = self.irecv.device_info['ECID']
        logging.debug(f'ECID: {ecid}')

        logging.info('device booted into recovery')

        # now we load the iBEC
        try:
            self.recovery_send_ibec()
        except USBError:
            pass

        self.reconnect()

        # now finally do the magic to put the device into restore mode
        self.recovery_enter_restore()

    def restore_device(self):
        # TODO: implement
        raise NotImplementedError('restore_device not yet implemented')

    def upgrade(self):
        self.boot_ramdisk()

        # device is finally in restore mode, let's do this
        self.restore_device()



import logging

from tqdm import tqdm
import click

from pymobiledevice3.cli.cli_common import Command
from pymobiledevice3.services.mobilebackup2 import Mobilebackup2Service
from pymobiledevice3.lockdown import LockdownClient


@click.group()
def cli():
    """ backup cli """
    pass


@cli.group()
def backup2():
    """ backup utils """
    pass


@backup2.command(cls=Command)
@click.argument('backup-directory', type=click.Path(exists=True, file_okay=False))
@click.option('--full', is_flag=True)
def backup(lockdown: LockdownClient, backup_directory, full):
    backup_client = Mobilebackup2Service(lockdown)
    with tqdm(total=100) as pbar:
        def update_bar(percentage):
            pbar.n = percentage
            pbar.refresh()

        backup_client.backup(full=full, backup_directory=backup_directory, progress_callback=update_bar)


@backup2.command(cls=Command)
@click.argument('backup-directory', type=click.Path(exists=True, file_okay=False))
@click.option('--system/--no-system', default=False)
@click.option('--reboot/--no-reboot', default=True)
@click.option('--copy/--no-copy', default=True)
@click.option('--settings/--no-settings', default=True)
@click.option('--remove/--no-remove', default=False)
@click.option('-p', '--password', type=click.STRING, default='')
def restore(lockdown: LockdownClient, backup_directory, system, reboot, copy, settings, remove, password):
    backup_client = Mobilebackup2Service(lockdown)
    with tqdm(total=100) as pbar:
        def update_bar(percentage):
            pbar.n = percentage
            pbar.refresh()

        backup_client.restore(backup_directory=backup_directory, progress_callback=update_bar, system=system,
                              reboot=reboot, copy=copy, settings=settings, remove=remove, password=password)


@backup2.command(cls=Command)
@click.argument('backup-directory', type=click.Path(exists=True, file_okay=False))
def info(lockdown: LockdownClient, backup_directory):
    backup_client = Mobilebackup2Service(lockdown)
    print(backup_client.info(backup_directory=backup_directory))


@backup2.command('list', cls=Command)
@click.argument('backup-directory', type=click.Path(exists=True, file_okay=False))
def list_(lockdown: LockdownClient, backup_directory):
    backup_client = Mobilebackup2Service(lockdown)
    print(backup_client.list(backup_directory=backup_directory))


@backup2.command(cls=Command)
@click.argument('backup-directory', type=click.Path(exists=True, file_okay=False))
@click.option('-p', '--password', type=click.STRING, default='')
def unback(lockdown: LockdownClient, backup_directory, password):
    backup_client = Mobilebackup2Service(lockdown)
    backup_client.unback(backup_directory=backup_directory, password=password)


@backup2.command(cls=Command)
@click.argument('domain-name', type=click.STRING)
@click.argument('relative-path', type=click.STRING)
@click.argument('backup-directory', type=click.Path(exists=True, file_okay=False))
@click.option('-p', '--password', type=click.STRING, default='')
def extract(lockdown: LockdownClient, domain_name, relative_path, backup_directory, password):
    backup_client = Mobilebackup2Service(lockdown)
    backup_client.extract(domain_name, relative_path, backup_directory=backup_directory, password=password)


@backup2.command(cls=Command)
@click.argument('on', type=click.BOOL)
@click.argument('password', type=click.STRING)
@click.argument('backup-directory', type=click.Path(exists=True, file_okay=False))
def encryption(lockdown: LockdownClient, backup_directory, on, password):
    will_encrypt = lockdown.get_value('com.apple.mobile.backup', 'WillEncrypt')
    if will_encrypt == on:
        logging.error('Encryption already ' + ('on!' if on else 'off!'))
        return
    backup_client = Mobilebackup2Service(lockdown)
    if on:
        backup_client.change_password(backup_directory, new=password)
    else:
        backup_client.change_password(backup_directory, old=password)


@backup2.command(cls=Command)
@click.argument('old-password', type=click.STRING)
@click.argument('new-password', type=click.STRING)
@click.argument('backup-directory', type=click.Path(exists=True, file_okay=False))
def change_password(lockdown: LockdownClient, old_password, new_password, backup_directory):
    will_encrypt = lockdown.get_value('com.apple.mobile.backup', 'WillEncrypt')
    if not will_encrypt:
        logging.error('Encryption is not turned on!')
        return
    backup_client = Mobilebackup2Service(lockdown)
    backup_client.change_password(backup_directory, old=old_password, new=new_password)


@backup2.command(cls=Command)
@click.argument('backup-directory', type=click.Path(exists=True, file_okay=False))
def erase_device(lockdown: LockdownClient, backup_directory):
    backup_client = Mobilebackup2Service(lockdown)
    backup_client.erase_device(backup_directory)



import tempfile
from pprint import pprint

import click

from pymobiledevice3.cli.cli_common import Command, print_json
from pymobiledevice3.tcp_forwarder import TcpForwarder


@click.group()
def cli():
    """ apps cli """
    pass


@cli.group('lockdown')
def lockdown_group():
    """ lockdown options """
    pass


@lockdown_group.command('forward', cls=Command)
@click.argument('src_port', type=click.IntRange(1, 0xffff))
@click.argument('dst_port', type=click.IntRange(1, 0xffff))
@click.option('-d', '--daemonize', is_flag=True)
def lockdown_forward(lockdown, src_port, dst_port, daemonize):
    """ forward tcp port """
    forwarder = TcpForwarder(lockdown, src_port, dst_port)

    if daemonize:
        try:
            from daemonize import Daemonize
        except ImportError:
            raise NotImplementedError('daemonizing is only supported on unix platforms')

        with tempfile.NamedTemporaryFile('wt') as pid_file:
            daemon = Daemonize(app=f'forwarder {src_port}->{dst_port}', pid=pid_file.name, action=forwarder.start)
            daemon.start()
    else:
        forwarder.start()


@lockdown_group.command('recovery', cls=Command)
def lockdown_recovery(lockdown):
    """ enter recovery """
    pprint(lockdown.enter_recovery())


@lockdown_group.command('service', cls=Command)
@click.argument('service_name')
def lockdown_service(lockdown, service_name):
    """ send-receive raw service messages """
    lockdown.start_service(service_name).shell()


@lockdown_group.command('info', cls=Command)
@click.option('--color/--no-color', default=True)
def lockdown_info(lockdown, color):
    """ query all lockdown values """
    print_json(lockdown.all_values, colored=color)


@lockdown_group.command('unpair', cls=Command)
def lockdown_unpair(lockdown):
    """ unpair from connected device """
    lockdown.unpair()



import time

from tqdm import trange
from usb.core import find, Device, USBError

USB_TIMEOUT = 10000

IRECV_K_RECOVERY_MODE_1 = 0x1280
IRECV_K_RECOVERY_MODE_2 = 0x1281
IRECV_K_RECOVERY_MODE_3 = 0x1282
IRECV_K_RECOVERY_MODE_4 = 0x1283
IRECV_K_WTF_MODE = 0x1222
IRECV_K_DFU_MODE = 0x122

APPLE_VENDOR_ID = 0x05AC


class IRecv:
    def __init__(self, ecid=None, timeout=0xffffffff):
        self._device = None  # type: Device
        self.device_info = {}
        self._find(ecid=ecid, timeout=timeout)
        self._populate_device_info()

    def ctrl_transfer(self, bmRequestType, bRequest):
        self._device.ctrl_transfer(bmRequestType, bRequest)

    def send_buffer(self, buf: bytes):
        packet_size = 0x8000

        # initiate transfer
        self.ctrl_transfer(0x41, 0)

        for i in trange(0, len(buf), packet_size):
            # Use bulk transfer for recovery mode and control transfer for DFU and WTF mode
            chunk = buf[i:i + packet_size]
            n = self._device.write(0x04, chunk, timeout=USB_TIMEOUT)
            if n != len(chunk):
                raise IOError('failed to upload data')

    def send_command(self, cmd: str, timeout=USB_TIMEOUT):
        self._device.ctrl_transfer(0x40, 0, 0, 0, cmd.encode() + b'\0', timeout=timeout)

    def getenv(self, name):
        try:
            self.send_command(f'getenv {name}')
        except USBError:
            return None
        return bytearray(self._device.ctrl_transfer(0xc0, 0, 0, 0, 255))

    def set_autoboot(self, enable: bool):
        self.send_command(f'setenv auto-boot {str(enable).lower()}')
        self.send_command('saveenv')

    def reboot(self):
        try:
            self.send_command('reboot')
        except USBError:
            pass

    def _find(self, ecid=None, timeout=0xffffffff):
        start = time.time()
        end = start + timeout
        while (self._device is None) and (time.time() < end):
            for device in find(find_all=True):
                try:
                    if not device.manufacturer.startswith('Apple'):
                        continue
                    if 'Recovery Mode' not in device.product:
                        continue
                    if self._device is not None:
                        raise Exception('More then one connected device was found connected in recovery mode')
                    self._device = device
                    self._populate_device_info()

                    if ecid is not None:
                        found_ecid = int(self.device_info['ECID'], 16)
                        if found_ecid != ecid:
                            # wrong device - move on
                            self._device = None
                            continue
                except ValueError:
                    continue

    def _populate_device_info(self):
        for component in self._device.serial_number.split(' '):
            k, v = component.split(':')
            if k == 'SRNM' and '[' in v:
                # trim the `[]`
                v = v[1:-1]
            self.device_info[k] = v

    def __str__(self):
        return str(self.device_info)


def main():
    print(IRecv())


if __name__ == '__main__':
    main()



import logging
import plistlib
import traceback

import IPython
import click

from pymobiledevice3.cli.cli_common import Command, print_json
from pymobiledevice3.irecv import IRecv
from pymobiledevice3.restore.restore import RestoreService


def return_to_normal_mode():
    device = IRecv()
    device.set_autoboot(True)
    device.reboot()


@click.group()
def cli():
    """ cli """
    pass


@cli.group()
def restore():
    """ restore options """
    pass


@restore.command('shell')
def restore_shell():
    """ create an IPython shell for interacting with iBoot """
    device = IRecv()
    IPython.embed(
        user_ns={
            'device': device,
        })


@restore.command('enter', cls=Command)
def restore_enter(lockdown):
    """ enter Recovery mode """
    lockdown.enter_recovery()


@restore.command('exit')
def restore_exit():
    """ exit Recovery mode """
    return_to_normal_mode()


@restore.command('tss', cls=Command)
@click.argument('ipsw', type=click.File('rb'))
@click.argument('out', type=click.File('wb'), required=False)
@click.option('--color/--no-color', default=True)
def restore_tss(lockdown, ipsw, out, color):
    """ query SHSH blobs """
    tss = RestoreService(lockdown, ipsw).fetch_tss_record()
    if out:
        plistlib.dump(tss, out)
    print_json(tss, colored=color)


@restore.command('ramdisk', cls=Command)
@click.argument('ipsw', type=click.File('rb'))
@click.option('--tss', type=click.File('rb'))
def restore_ramdisk(lockdown, ipsw, tss):
    """ don't perform an actual restore. just enter the update ramdisk """
    try:
        if tss:
            tss = plistlib.load(tss)
        RestoreService(lockdown, ipsw, tss=tss).boot_ramdisk()
    except Exception:
        traceback.print_exc()
        logging.info('returning to normal mode')
        return_to_normal_mode()


@restore.command('upgrade', cls=Command)
@click.argument('ipsw', type=click.File('rb'))
@click.option('--tss', type=click.File('rb'))
def restore_upgrade(lockdown, ipsw, tss):
    """ perform an upgrade """
    try:
        if tss:
            tss = plistlib.load(tss)
        RestoreService(lockdown, ipsw, tss=tss).upgrade()
    except Exception:
        traceback.print_exc()
        logging.info('returning to normal mode')
        return_to_normal_mode()



#!/usr/bin/env python3
import logging

from pymobiledevice3.lockdown import LockdownClient


class PreboardService(object):
    SERVICE_NAME = 'com.apple.preboardservice_v2'

    def __init__(self, lockdown: LockdownClient):
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown
        self.c = self.lockdown.start_service(self.SERVICE_NAME)

    def create_stashbag(self, manifest):
        return self.c.send_recv_plist({'Command': 'CreateStashbag', 'Manifest': manifest})

    def commit(self, manifest):
        return self.c.send_recv_plist({'Command': 'CommitStashbag', 'Manifest': manifest})



#!/usr/bin/env python3
import logging
from pathlib import Path
import time
import uuid
import plistlib
from datetime import datetime
from contextlib import contextmanager

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.exceptions import PyMobileDevice3Exception, AfcFileNotFoundError, AfcException
from pymobiledevice3.services.afc import AfcService, AFC_LOCK_EX, afc_error_t, AFC_LOCK_UN
from pymobiledevice3.services.device_link import DeviceLink
from pymobiledevice3.services.installation_proxy import InstallationProxyService
from pymobiledevice3.services.notification_proxy import NotificationProxyService
from pymobiledevice3.services.springboard import SpringBoardServicesService

SUPPORTED_VERSIONS = [2.0, 2.1]
ITUNES_FILES = [
    'ApertureAlbumPrefs', 'IC-Info.sidb', 'IC-Info.sidv', 'PhotosFolderAlbums', 'PhotosFolderName',
    'PhotosFolderPrefs', 'VoiceMemos.plist', 'iPhotoAlbumPrefs', 'iTunesApplicationIDs', 'iTunesPrefs',
    'iTunesPrefs.plist'
]
NP_SYNC_WILL_START = 'com.apple.itunes-mobdev.syncWillStart'
NP_SYNC_DID_START = 'com.apple.itunes-mobdev.syncDidStart'
NP_SYNC_LOCK_REQUEST = 'com.apple.itunes-mobdev.syncLockRequest'
NP_SYNC_DID_FINISH = 'com.apple.itunes-mobdev.syncDidFinish'


class Mobilebackup2Service:
    SERVICE_NAME = 'com.apple.mobilebackup2'

    def __init__(self, lockdown: LockdownClient):
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown
        self.service = self.lockdown.start_service(self.SERVICE_NAME)

    def backup(self, full=True, backup_directory=Path('.'), progress_callback=lambda x: None):
        backup_directory = Path(backup_directory)
        device_directory = backup_directory / self.lockdown.identifier
        device_directory.mkdir(exist_ok=True, mode=0o755, parents=True)

        with self.device_link(backup_directory) as dl:
            notification_proxy = NotificationProxyService(self.lockdown)
            afc = AfcService(self.lockdown)

            with self._backup_lock(afc, notification_proxy):
                # Initialize Info.plist
                info_plist = self.init_mobile_backup_factory_info(afc)
                with open(device_directory / 'Info.plist', 'wb') as fd:
                    plistlib.dump(info_plist, fd)

                # Initialize Status.plist file if doesn't exist.
                status_path = device_directory / 'Status.plist'
                if full or not status_path.exists():
                    with open(device_directory / 'Status.plist', 'wb') as fd:
                        plistlib.dump({
                            'BackupState': 'new',
                            'Date': datetime.utcnow(),
                            'IsFullBackup': full,
                            'Version': '3.3',
                            'SnapshotState': 'finished',
                            'UUID': str(uuid.uuid4()).upper(),
                        }, fd, fmt=plistlib.FMT_BINARY)

                # Create Manifest.plist if doesn't exist.
                manifest_path = device_directory / 'Manifest.plist'
                if full:
                    manifest_path.unlink(missing_ok=True)
                (device_directory / 'Manifest.plist').touch()

                dl.send_process_message({'MessageName': 'Backup', 'TargetIdentifier': self.lockdown.identifier})
                dl.dl_loop(progress_callback)

    def restore(self, backup_directory=Path('.'), system=False, reboot=True, copy=True, settings=True, remove=False,
                password='', progress_callback=lambda x: None):
        backup_directory = Path(backup_directory)
        device_directory = backup_directory / self.lockdown.identifier
        assert device_directory.exists()

        with self.device_link(backup_directory) as dl:
            notification_proxy = NotificationProxyService(self.lockdown)
            afc = AfcService(self.lockdown)

            with self._backup_lock(afc, notification_proxy):
                assert (device_directory / 'Info.plist').exists()
                manifest_plist_path = device_directory / 'Manifest.plist'
                assert manifest_plist_path.exists()
                with open(manifest_plist_path, 'rb') as fd:
                    manifest = plistlib.load(fd)
                is_encrypted = manifest.get('IsEncrypted', False)
                options = {
                    'RestoreShouldReboot': reboot,
                    'RestoreDontCopyBackup': copy,
                    'RestorePreserveSettings': settings,
                    'RestoreSystemFiles': system,
                    'RemoveItemsNotRestored': remove,
                }
                if is_encrypted:
                    if password:
                        options['Password'] = password
                    else:
                        self.logger.error('Backup is encrypted, please supply password.')
                        return
                dl.send_process_message({
                    'MessageName': 'Restore',
                    'TargetIdentifier': self.lockdown.identifier,
                    'SourceIdentifier': self.lockdown.identifier,
                    'Options': options,
                })
                dl.dl_loop(progress_callback)

    def info(self, backup_directory=Path('.')):
        with self.device_link(Path(backup_directory)) as dl:
            dl.send_process_message({'MessageName': 'Info', 'TargetIdentifier': self.lockdown.identifier})
            result = dl.dl_loop()
        return result

    def list(self, backup_directory=Path('.')):
        with self.device_link(Path(backup_directory)) as dl:
            dl.send_process_message({
                'MessageName': 'List', 'TargetIdentifier': self.lockdown.identifier,
                'SourceIdentifier': self.lockdown.identifier
            })
            result = dl.dl_loop()
        return result

    def unback(self, backup_directory=Path('.'), password=''):
        device_directory = Path(backup_directory) / self.lockdown.identifier
        assert (device_directory / 'Info.plist').exists()
        assert (device_directory / 'Manifest.plist').exists()
        with self.device_link(Path(backup_directory)) as dl:
            message = {'MessageName': 'Unback', 'TargetIdentifier': self.lockdown.identifier}
            if password:
                message['Password'] = password
            dl.send_process_message(message)
            result = dl.dl_loop()
        return result

    def extract(self, domain_name, relative_path, backup_directory=Path('.'), password=''):
        device_directory = backup_directory / self.lockdown.identifier
        assert (device_directory / 'Info.plist').exists()
        assert (device_directory / 'Manifest.plist').exists()
        with self.device_link(Path(backup_directory)) as dl:
            message = {
                'MessageName': 'Extract', 'TargetIdentifier': self.lockdown.identifier, 'DomainName': domain_name,
                'RelativePath': relative_path
            }
            if password:
                message['Password'] = password
            dl.send_process_message(message)
            result = dl.dl_loop()
        return result

    def change_password(self, backup_directory=Path('.'), old='', new=''):
        with self.device_link(Path(backup_directory)) as dl:
            message = {'MessageName': 'ChangePassword', 'TargetIdentifier': self.lockdown.identifier}
            if old:
                message['OldPassword'] = old
            if new:
                message['NewPassword'] = new
            dl.send_process_message(message)
            dl.dl_loop()

    def erase_device(self, backup_directory=Path('.')):
        with self.device_link(Path(backup_directory)) as dl:
            dl.send_process_message({'MessageName': 'EraseDevice', 'TargetIdentifier': self.lockdown.identifier})
            dl.dl_loop()

    def version_exchange(self, dl: DeviceLink, local_versions=None):
        if local_versions is None:
            local_versions = SUPPORTED_VERSIONS
        dl.send_process_message({
            'MessageName': 'Hello',
            'SupportedProtocolVersions': local_versions,
        })
        reply = dl.receive_message()
        assert reply[0] == 'DLMessageProcessMessage' and reply[1]['ErrorCode'] == 0
        assert reply[1]['ProtocolVersion'] in SUPPORTED_VERSIONS

    def init_mobile_backup_factory_info(self, afc: AfcService):
        ip = InstallationProxyService(self.lockdown)
        sbs = SpringBoardServicesService(self.lockdown)

        root_node = self.lockdown.get_value()
        itunes_settings = self.lockdown.get_value(domain='com.apple.iTunes')
        min_itunes_version = self.lockdown.get_value('com.apple.mobile.iTunes', 'MinITunesVersion')
        app_dict = {}
        installed_apps = []
        apps = ip.browse(options={'ApplicationType': 'User'},
                         attributes=['CFBundleIdentifier', 'ApplicationSINF', 'iTunesMetadata'])
        for app in apps:
            bundle_id = app['CFBundleIdentifier']
            if bundle_id:
                installed_apps.append(bundle_id)
                if app.get('iTunesMetadata', False) and app.get('ApplicationSINF', False):
                    app_dict[bundle_id] = {
                        'ApplicationSINF': app['ApplicationSINF'],
                        'iTunesMetadata': app['iTunesMetadata'],
                        'PlaceholderIcon': sbs.get_icon_pngdata(bundle_id),
                    }

        files = {}
        for file in ITUNES_FILES:
            try:
                data_buf = afc.get_file_contents('/iTunes_Control/iTunes/' + file)
            except AfcFileNotFoundError:
                pass
            else:
                files[file] = data_buf

        ret = {
            'iTunes Version': min_itunes_version if min_itunes_version else '10.0.1',
            'iTunes Files': files,
            'Unique Identifier': self.lockdown.identifier.upper(),
            'Target Type': 'Device',
            'Target Identifier': root_node['UniqueDeviceID'],
            'Serial Number': root_node['SerialNumber'],
            'Product Version': root_node['ProductVersion'],
            'Product Type': root_node['ProductType'],
            'Installed Applications': installed_apps,
            'GUID': uuid.uuid4().bytes,
            'Display Name': root_node['DeviceName'],
            'Device Name': root_node['DeviceName'],
            'Build Version': root_node['BuildVersion'],
            'Applications': app_dict,
        }

        if 'IntegratedCircuitCardIdentity' in root_node:
            ret['ICCID'] = root_node['IntegratedCircuitCardIdentity']
        if 'InternationalMobileEquipmentIdentity' in root_node:
            ret['IMEI'] = root_node['InternationalMobileEquipmentIdentity']
        if 'MobileEquipmentIdentifier' in root_node:
            ret['MEID'] = root_node['MobileEquipmentIdentifier']
        if 'PhoneNumber' in root_node:
            ret['Phone Number'] = root_node['PhoneNumber']

        try:
            data_buf = afc.get_file_contents('/Books/iBooksData2.plist')
        except AfcFileNotFoundError:
            pass
        else:
            ret['iBooks Data 2'] = data_buf
        if itunes_settings:
            ret['iTunes Settings'] = itunes_settings
        return ret

    @contextmanager
    def _backup_lock(self, afc, notification_proxy):
        notification_proxy.notify_post(NP_SYNC_WILL_START)
        lockfile = afc.fopen('/com.apple.itunes.lock_sync', 'r+')
        if lockfile:
            notification_proxy.notify_post(NP_SYNC_LOCK_REQUEST)
            for _ in range(50):
                try:
                    afc.lock(lockfile, AFC_LOCK_EX)
                except AfcException as e:
                    if e.status == afc_error_t.OP_WOULD_BLOCK:
                        time.sleep(0.2)
                    else:
                        afc.fclose(lockfile)
                        raise
                else:
                    notification_proxy.notify_post(NP_SYNC_DID_START)
                    break
            else:  # No break, lock failed.
                afc.fclose(lockfile)
                raise PyMobileDevice3Exception('Failed to lock itunes sync file')
        try:
            yield
        finally:
            afc.lock(lockfile, AFC_LOCK_UN)
            afc.fclose(lockfile)
            notification_proxy.notify_post(NP_SYNC_DID_FINISH)

    @contextmanager
    def device_link(self, backup_directory):
        dl = DeviceLink(self.service, backup_directory)
        dl.version_exchange()
        self.version_exchange(dl)
        try:
            yield dl
        finally:
            dl.disconnect()
